kong-plugin-enterprise-exit-transformer
=======================================

This plugin allows for customizing kong response exit messages through lua
functions. This allows from changing the structure of the message, to
changing their status codes, headers and messages.

**Note** `KONG_UNTRUSTED_LUA` must be set to either `on` or `sandbox` for this
plugin to work.

## Usage

### Add a service
```bash
http :8001/services name=example.com host=mockbin.org
http -f :8001/services/example.com/routes hosts=example.com
```

### Create some transformer lua code

Exit transformer exepects some lua code implementing a function that receives
`status`, `body` and `headers`, and returns `status`, `body` and `headers`.

```lua
return function (status, body, headers)
  if not body or not body.message then
    return status, body, headers
  end

  -- Let's say we want to transform errors so they are always a 200, with a
  -- very particular syntax.

  body = {
    ["kong_code"] = status,
    ["kong_message"] = body.message,
    ["error"] = true,
  }

  status = 200

  return status, body, headers
end
```

### Configure the plugin

```bash
http -f :8001/services/example.com/plugins name=exit-transformer config.functions=@some-code.lua
```

### Enjoy your new error messages

```bash
# Enable key auth
http :8001/services/example.com/plugins name=key-auth
# Make a request without key auth
$ http :8000/request/foo Host:example.com
HTTP/1.1 200 OK
Connection: keep-alive
Content-Length: 75
Content-Type: application/json; charset=utf-8
Date: Mon, 19 Aug 2019 19:15:43 GMT
Server: kong/0.36-dev-enterprise-edition
WWW-Authenticate: Key realm="kong"

{
    "error": true,
    "kong_code": 401,
    "kong_message": "No API key found in request"
}
```

## Configuration

* `config.functions`: list of lua functions to reduce exit calls into.
* `config.handle_unknown`: handle calls that are not tied to any service or
   route or workspace, ie: a 404 on the proxy side. **NOTE** This flag will
   only apply to one configuration entry.
* `config.handle_unexpected`: handle unexpected errors from Kong.

## Examples

Following some examples showcasing exit-transformer plugin to customize
Kong error responses

### Custom JSON errors for Kong

[custom-json-error.lua](examples/custom-json-error.lua).

This example shows a case where a customer might want Kong to return a
particular involved JSON response for errors. Useful for integrating Kong with
a legacy system that requires a particular response syntax.

```
http -f :8001/plugins name=exit-transformer \
                      config.handle_unknown=true \
                      config.handle_unexpected=true \
                      config.functions=@examples/custom-json-error.lua
```


### Custom errors by mimetype (json, html)

[custom-errors-by-mimetype.lua](examples/custom-errors-by-mimetype.lua).

This example shows a case where a customer wants custom JSON and HTML responses
based on an `Accept` header. Usually when a customer doesn't want public
error responses to contain the word "kong".

This example requires the following settings:

```
KONG_UNTRUSTED_LUA_SANDBOX_REQUIRES = resty.template,kong.tools.utils
KONG_UNTRUSTED_LUA_SANDBOX_ENVIRONMENT = kong.request
```

```
http -f :8001/plugins name=exit-transformer \
                      config.handle_unknown=true \
                      config.handle_unexpected=true \
                      config.functions=@examples/custom-errors-by-mimetype.lua
```

<p align="center">
  <img src="https://user-images.githubusercontent.com/208952/89324053-edb14180-d686-11ea-9a20-dfcf336d7746.png" width="500px">
</p>
