[![Build Status][badge-travis-image]][badge-travis-url]

# Kong Response Transformer Advanced Plugin

Transform the response sent by the upstream server before returning the response to the client.

This plugin is a superset of the Response Transformer plugin included in Kong
and provides additional functionality:
* When transforming a JSON payload, transformations are applied to nested JSON objects and
arrays. This can be turned off using the `config.dots_in_keys` configuration parameter.
* Transformations can be restricted to responses with specific status codes using various
`config.*.if_status` configuration parameters.
* JSON body contents can be restricted to a set of allowed properties with
`config.allow.json`.
* The entire response body can be replaced using `config.replace.body`.
* Arbitrary transformation functions written in Lua can be applied.

## Status

This plugin is bundled with Kong Enterprise as of 0.35.

## Configuration

### Parameters

Configuration parameters below are inherited from Kong Response Transformer:

| Form parameter | default | description
|----------------|---------|------------
| `name`         || The name of the plugin to configure; in this case, `response-transformer-advanced`
| `config.remove.headers` || List of header names. Unset the header(s) with the given name
| `config.remove.json` || List of property names. Remove the property from the JSON body if it is present
| `config.rename.headers` || List of header name `name1:name2` pairs. If a header with `name1` exists and `name2` is valid, rename header with `name2`.
| `config.replace.headers` || List of header `name:value` pairs. If and only if the header is already set, replace its old value with the new one. Ignored if the header is not already set
| `config.replace.json` || List of property `name:value` pairs. If and only if the parameter is already present, replace its old value with the new one. Ignored if the parameter is not already present
| `config.replace.json_types` || List of JSON type names. Specify the types of the JSON values returned when replacing JSON properties.
| `config.add.headers` || List of headername:value pairs. If and only if the header is not already set, set a new header with the given value. Ignored if the header is already set
| `config.add.json` || List of property:value pairs. If and only if the property is not present, add a new property with the given value to the JSON body. Ignored if the property is already present
| `config.add.json_types` || List of JSON type names. Specify the types of the JSON values returned when adding a new JSON property.
| `config.append.headers` || List of headername:value pairs. If the header is not set, set it with the given value. If it is already set, a new header with the same name and the new value will be set
| `config.append.json` || List of property:value pairs. If the property is not present in the JSON body, add it with the given value. If it is already present, the two values (old and new) will be aggregated in an array
| `config.replace.json_types` || List of JSON type names. Specify the types of the JSON values returned when appending JSON properties.

New configuration directives:

| Form parameter | default | description
|----------------|---------|------------
| `config.add.if_status` || List of response status codes or ranges to which the transformation will apply. Empty means all response codes
| `config.append.if_status` || List of response status codes or ranges to which the transformation will apply. Empty means all response codes
| `config.remove.if_status` || List of response status codes or ranges to which the transformation will apply. Empty means all response codes
| `config.replace.if_status` || List of response status codes or ranges to which the transformation will apply. Empty means all response codes
| `config.transform.functions` || Set of lua functions to perform arbitrary transforms in a response JSON body
| `config.transform.if_status` || List of response status codes or ranges to which the arbitrary transformation will apply. Empty means all response codes
| `config.transform.json` || Apply lua functions to a particular list of JSON property `name` or `name:value` pairs.
| `config.replace.body` || String with which to replace the entire response body
| `config.allow.json` || Set of parameter names. Only allows accepted parameters in the JSON response body.
| `config.dots_in_keys` | false | Whether dots should be treated as part of a property name or used to descend into nested JSON objects.

### Examples

- Add multiple headers by passing each header:value pair separately:

```
$ curl -X POST http://localhost:8001/routes/{route id}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.add.headers[1]=h1:v1" \
  --data "config.add.headers[2]=h2:v1"
```

- Add the headers, as above, but only if the response code is 200

```
$ curl -X POST http://localhost:8001/routes/{route id}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.add.headers[1]=h1:v1" \
  --data "config.add.headers[2]=h2:v1" \
  --data "config.add.if_status=200"
```

- Replace entire response body if response code is 500

```
$ curl -X POST http://localhost:8001/routes/{route id}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.replace.body='{\"error\": \"internal server error\"}'" \
  --data "config.replace.if_status=500"
```

- Explicitly specify the types of added JSON values if the response code is 500

```
$ curl -X POST http://localhost:8001/routes/{route id}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.add.json_types=number" \
  --data "config.replace.if_status=500"
```

- Remove nested JSON content:

```
$ curl -X POST http://localhost:8001/routes/{route id}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.remove.json=customers.info.ssn"
```

**Note**: the plugin doesn't validate the value in `config.replace.body` against
the content type as defined in the `Content-Type` response header.

- Perform arbitrary transforms on a response body

Given some lua code in a file called wrangler.lua

```lua
--- wrangler.lua
return function (data)
  if type(data) ~= "table" then
    return data
  end

  -- remove foo key
  data["foo"] = nil

  -- add a new key
  data["foobar"] = "hello world"

  return data
end
```

And the plugin configured on a service

```
$ curl -X POST http://localhost:8001/services/{service}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.transform.functions=@wrangler.lua
```

Given this response

```json
{
  "foo": "bar",
  "something": "else"
}
```

The response will be transformed into

```json
{
  "something": "else",
  "foobar": "hello world"
}
```

Arbitrary transforms can be applied to a set of JSON property names, like

```lua
-- universe.lua
-- answers a question only when its known
return function (data, key)
  if data[key] == "meaning of the universe" then
    data["a"] = 42
  end
end
```

And the plugin configured on a service

```
$ curl -X POST http://localhost:8001/services/{service}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.transform.functions=@universe.lua
  --data "config.transform.json[1]=questions[*].q"
```

Given this response

```json
{
  "questions": [
    { "q": "knock, knock" },
    { "q": "meaning of the universe" }
  ]
}
```

The response will be transformed into

```json
{
  "questions": [
    { "q": "knock, knock" },
    { "q": "meaning of the universe", "a": 42}
  ]
}
```

An additional parameter can be passed to the transform function by using
`name:value` pairs on `config.transform.json`. Following a similar example

```lua
-- answers a question depending on value
return function (data, key, value)
  if data[key] == value then
    data["a"] = 42
  end
end
```

And the plugin configured on a service

```
$ curl -X POST http://localhost:8001/services/{service}/plugins \
  --data "name=response-transformer-advanced" \
  --data "config.transform.functions=@universe.lua
  --data "config.transform.json[1]=questions[*].q:meaning of the universe"
  --data "config.transform.json[2]=questions[*].q:meaning of everything"
```

Given this response

```json
{
  "questions": [
    { "q": "knock, knock" },
    { "q": "meaning of the universe" },
    { "q": "meaning of everything" }
  ]
}
```

The response will be transformed into

```json
{
  "questions": [
    { "q": "knock, knock" },
    { "q": "meaning of the universe", "a": 42}
    { "q": "meaning of everything", "a": 42}
  ]
}
```


[badge-travis-url]: https://travis-ci.com/Kong/kong-plugin-response-transformer-advanced/branches
[badge-travis-image]: https://travis-ci.com/Kong/kong-plugin-response-transformer-advanced.svg?token=BfzyBZDa3icGPsKGmBHb&branch=master
