# openapi3-deserializer
Library to deserialize OAS3 type parameters

Specs: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#parameter-object

Examples: https://swagger.io/docs/specification/serialization/


The library only does deserialization for now, to be able to parse parameters
to validate them.


## Basics

The module returns a single function used for deserialization

```lua
local deserialize = require "resty.openapi3.deserializer"

local p_type = "array"
local p_style = "matrix"
local p_explode = true
local value = ";id=3;id=4;id=5"

local result, err = deserialize(p_style, p_type, p_explode, value)

assert.same( nil, err )
assert.same({ 3, 4, 5 }, result )
```

### Parameter types:

- `primitive`
- `array`
- `object`

_Note_: `p_type` must be a string, but any value passed other than `"object"` or
`"array"` will be substituted for `"primitive"`. This allows for easy integration
with JSONschema, since you can pass `schema.type` as the value for `p_type`.

### Serialization styles:

- `matrix` (primitive, array, object)
- `label` (primitive, array, object)
- `form` (primitive, array, object)
- `simple` (primitive, array, object)
- `spaceDelimited` (array)
- `pipeDelimited` (array)
- `deepObject` (object)

## Testing

- lint using luacheck
- tests using busted

validate everything by running: `busted && luacheck .`


## Caveats

- errors are only thrown for bad input (eg. non-existing styles/types). Any
  deserialization error is returned as `nil+err`

- for an object passed as query parameters, in style `form` with `explode=true`
  the entire query_args table will be the object, instead of a single value
  from that table. You can pass that table
  as the value, and it will be returned without any processing.

- for `deepObject` the signature is different. It is only supported for query
  arguments. It needs the full list of query args, and the name of the
  parameter to be deserialized.

  ```lua
  local result, err = deserialize("deepObject", "object", true, param_name, query_args)
  ```

  This may return `nil, "not found"` if the object does not exist.

## History

Versioning is strictly SemVer based

### 1.1.0 add normalization for decoder

 - Fix: correctly decode and normalize array when multiple headers with 
   the same field-name
 - Tests: extend tests for header->primitive, header->array and header->object cases

### 1.0.1 adding copyright header

 - Chore: Add copyright header to files

### 1.0.0 production ready release, 15-Jul-2020

 - Fix: Minor improvement in error messages
 - Fix: if a query argument is parsed as an array, but has a single value, then
   OR would not provide a table, but only a single string value. Which would
   cause a decoded array with a return value of a string.

### 0.1.0 first release, 29-Jul-2019

 - initial release
