-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

--[[
    A Query AST is useful to manage the raw collection of
    nodes with DocumentNode at the top. It allows analysis and processing
    of data regarding the tree without modifying the original raw output tree
    rooted at the Document Node. Its functionality includes data decoration,
    generating new contexts within the raw tree, and utility routines that
    only make sense in the tree context such as union, difference, etc.
  ]]
local FragmentSpread = require"kong.gql.query.nodes.fragment_spread"
local FieldNode = require"kong.gql.query.nodes.field_node"
local InlineFragmentNode = require"kong.gql.query.nodes.inline_fragment_node"
local OperationNode = require"kong.gql.query.nodes.operation_node"
local DocumentNode = require"kong.gql.query.nodes.document_node"
local FragmentDefNode = require"kong.gql.query.nodes.fragment_def_node"

local IterStack = require"kong.gql.query.build_ast.iter_stack"
local util = require"kong.gql.util"

local _QueryAst = {}


function _QueryAst.new(document_node)
    local self = {}
    self.root = document_node
    self.data = {}
    _QueryAst.__index = _QueryAst
    setmetatable(self, _QueryAst)

    -- Resolve fragment spread
    self:_resolve_fragment_spread(self.root)

    return self
end


-- Resolves fragments naturally
-- @param node: a node not a document node
function _QueryAst:get_children_iter(node)
    local iter_s = IterStack.new()
    iter_s:push(self:get_raw_children_iter(node))

    local next
    local function get_next_child()
        next = iter_s:get_next()
        if not next then
            return nil
        end

        if next:is(InlineFragmentNode) or next:is(FragmentSpread) then
            iter_s:push(util.lst_to_iter(next.selection_set))
            return get_next_child()
        end

        return next
    end

    return get_next_child
end


function _QueryAst:get_raw_children_iter(node)
    if node:is(DocumentNode) then
        return util.lst_to_iter(node.definitions)
    elseif node:is(FieldNode) or node:is(OperationNode) or node:is(FragmentDefNode)
            or node:is(InlineFragmentNode) then
        return util.lst_to_iter(node.selection_set)
    elseif node:is(FragmentSpread) then
        return nil
    end
end


-- Entry point to use _resolve_from_frag_def recursively on all query nodes
function _QueryAst:_resolve_fragment_spread(node)
    if node:is(FragmentSpread) then
        node.selection_set = self:_resolve_from_frag_def(node)
        return
    end

    for child in _QueryAst:get_raw_children_iter(node) do
        self:_resolve_fragment_spread(child)
    end
end


-- Clones fragment definition, including nested fragment definitions
-- @param node: entry point will be FragmentSpread, then anything in subtree will be in this context
-- @return table, a list of resolved children subtrees, i.e. nodes with children and so on
function _QueryAst:_resolve_from_frag_def(node)
    local r = {}

    local src
    if node:is(FragmentSpread) then
        node:link_def(self.root:get_frag_def(node.name))
        src = node.def
    else src = node end

    for child in self:get_raw_children_iter(src) do
        if child:is(FragmentSpread) then
            child.selection_set = self:_resolve_from_frag_def(child)
            table.insert(r, child)
        else
            local o = {} -- New context with same attributes
            child.__index = child
            setmetatable(o, child)
            o.selection_set = self:_resolve_from_frag_def(o)
            table.insert(r, o)
        end
    end

    return r
end


function _QueryAst:get_data(node, key)
    local d = self.data[node]
    if d ~= nil then return d[key] end
end


function _QueryAst:set_data(node, key, value)
    local d = self.data[node]
    if not d then
        self.data[node] = {}
        self.data[node][key] = value
        return nil
    end

    d[key] = value
    return true
end


function _QueryAst:_inject_data(node, data)
    for k, v in pairs(data) do
        self:set_data(node, k, v)
    end
end


function _QueryAst:__decorate_data(node, type_ctx, data)
    if type_ctx:has_ref() then
        self:_inject_data(node, data)
    end

    for child in self:get_children_iter(node) do
        local child_fname = self:get_field_name(child)
        local new_ctx = type_ctx:next_field(child_fname)

        self:__decorate_data(child, new_ctx, data)
    end
end


function _QueryAst:decorate_data(types, schema, data)
    local type_ctx = schema:get_type_ctx_o(types)

    local op_node = self:get_op_node()
    local op_name = self:get_field_name(op_node)
    type_ctx = type_ctx:next_field(op_name)
    self:__decorate_data(op_node, type_ctx, data)
end


function _QueryAst:get_argument(node, key)
    if not node:is(FieldNode) then
        return nil
    end
    return node.arguments[key]
end


function _QueryAst:get_field_name(node)
    if node:is(OperationNode) then
        return node.type
    elseif node:is(FieldNode) then
        return node.name
    end
end


function _QueryAst:get_op_node()
    return self.root:get_op_node()
end


return _QueryAst
