-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local EnumTypeDef = require"kong.gql.schema.typedefs.enum"
local InputTypeDef = require"kong.gql.schema.typedefs.input"
local InterfaceTypeDef = require"kong.gql.schema.typedefs.interface"
local ObjectTypeDef = require"kong.gql.schema.typedefs.object"
local ScalarTypeDef = require"kong.gql.schema.typedefs.scalar"
local UnionTypeDef = require"kong.gql.schema.typedefs.union"
local SchemaTypeDef = require"kong.gql.schema.typedefs.schema"
local FieldDef = require"kong.gql.schema.field"

local Type = require"kong.gql.schema.type"
local Tdn = require"kong.gql.schema.tdn"

local util = require"kong.gql.util"
local lst_to_iter = util.lst_to_iter
local lst_to_map = util.lst_to_map


local _SchemaAst = {}


local TYPE_INTROSPECTION_QUERY = [[
fragment TypeAttr on __Type {
        kind,
        name
    }

# Wrapped type only goes max five levels
fragment WrappedTypeRef on __Type {
    ...TypeAttr
    ofType {
        ...TypeAttr,
        ofType {
            ...TypeAttr,
            ofType {
                ...TypeAttr,
                ofType {
                    ...TypeAttr
                }
            }
        }
    }
}

query {
    __schema {
        types {
             ...WrappedTypeRef

            # OBJECT and INTERFACE
            fields {
                name,
                args {
                    name,

                    type {
                        ...WrappedTypeRef
                    }
                }

                type {
                    ...WrappedTypeRef
                }
            }

            # Interfaces that OBJECT implements
            interfaces {
                name
            }

            # For Unions and Interfaces
            possibleTypes {
                ...WrappedTypeRef
            }

            # For Enums
            enumValues {
                name
            }
        }
    }
}
]]


function _SchemaAst.new(schema_def)
    local self = {}
    self.schema_def = schema_def
    self.data = {}
    setmetatable(self, _SchemaAst)
    _SchemaAst.__index = _SchemaAst
    return self
end


function _SchemaAst:get_typedef(type)
    return self.schema_def.typedefs[type]
end


function _SchemaAst:get_field_type(typedef, field_name)
    return typedef.fields[field_name].kind
end


local _TypeCtxObj = {}


-- @param typedef
function _TypeCtxObj.new(tdn, schema, typedef)
    local self = {}
    self.tdn = tdn
    self.schema = schema
    self.c_typedef = typedef

    _TypeCtxObj.__index = _TypeCtxObj
    return setmetatable(self, _TypeCtxObj)
end


local ROOT_OP_FIELD_TO_TYPE_NAME = {
    query = "Query",
    mutation = "Mutation",
    subscription = "Subscription"
}


function _TypeCtxObj:next_field(field_name)
    if not self.c_typedef then -- Must be root operation
        local root_op_typename = ROOT_OP_FIELD_TO_TYPE_NAME[field_name]
        return _TypeCtxObj.new(
                self.tdn:next_token(root_op_typename),
                self.schema,
                self.schema:get_typedef(root_op_typename)
        )
    end

    if not self.c_typedef:is(ObjectTypeDef) and
            not self.c_typedef:is(InterfaceTypeDef) then
        return error("Field base type must be object or interface")
    end

    local field_type = self.schema:get_field_type(self.c_typedef, field_name)
    if not field_type then
        local error_str = string.format(
                "Field selection `%s` of type `%s` does not exist based on schema",
                field_name,
                self.c_typedef.name
        )
        return error(error_str)
    end

    local base_field_type = field_type:unwrap_f()
    local base_field_typename = base_field_type:name()
    return _TypeCtxObj.new(
            self.tdn:next_token(base_field_typename, field_name),
            self.schema,
            self.schema:get_typedef(base_field_typename)
    )
end


-- TODO: Refactor some Tdn functionality into TypeCtxObj
function _TypeCtxObj:has_ref()
    return self.tdn:should_decorate()
end


function _SchemaAst:get_type_ctx_o(tdn_strings)
    local tdn = Tdn.new(tdn_strings)
    return _TypeCtxObj.new(tdn, self)
end


local function _djson_type(t)
    if type(t.ofType) ~= "table" then -- non-wrapped type
        return Type.new(t.name)
    end

    local wrapper_directive = Type.WRAPPER[t.kind]
    return Type.new(
            _djson_type(t.ofType),
            wrapper_directive
    )
end


local function _djson_fields(fields)
    if not fields then return nil end
    local res = {}
    for field in lst_to_iter(fields) do
        local field_def = FieldDef.new(
                field.name,
                _djson_type(field.type)
        )
        table.insert(res, field_def)
    end

    return lst_to_map(
            res,
            function (field_def)
                return field_def.name
            end
    )
end


local function _djson_implements(implements)
    if not implements then return nil end
    local res = {}
    for interface in lst_to_iter(implements) do
        table.insert(res, _djson_type(interface))
    end

    return res
end


local function _djson_enum_values(values)
    local res = {}
    for val in lst_to_iter(values) do
        table.insert(res, val.name)
    end

    return res
end


local function _djson_typedef(typedef)
    if typedef.kind == "OBJECT" and typedef.interfaces ~= ngx.null then
        local implements = _djson_implements(typedef.interfaces)
        return ObjectTypeDef(
                typedef.name,
                _djson_fields(typedef.fields),
                implements
        )
    elseif typedef.kind == "INTERFACE" and typedef.fields ~= ngx.null then

        local fields = _djson_fields(typedef.fields)
        return InterfaceTypeDef(
                typedef.name,
                fields,
                _djson_implements(typedef.possibleTypes)
        )
    elseif typedef.kind == "UNION" and typedef.possibleTypes ~= ngx.null then
        return UnionTypeDef(
                typedef.name,
                _djson_implements(typedef.possibleTypes)
        )
    elseif typedef.kind == "SCALAR" and typedef.name then
        return ScalarTypeDef(typedef.name)
    elseif typedef.kind == "INPUT_OBJECT" and typedef.fields ~= ngx.null then
        return InputTypeDef(
                typedef.name,
                _djson_fields(typedef.fields)
    )
    elseif typedef.kind == "ENUM" and typedef.enumValues ~= ngx.null then
        local enum_values = _djson_enum_values(typedef.enumValues)
        return EnumTypeDef(
                typedef.name,
                enum_values
        )
    else
        return nil, "error: TypeDef kind incorrect"
    end
end


-- @param json_data: table, entry point __schema key
function _SchemaAst.deserialize_json_data(json_data)
    local schema_data = json_data["__schema"]
    local typedefs = {}
    for typedef in lst_to_iter(schema_data["types"]) do
        if string.find(typedef.name, "__") ~= 1 then
            typedefs[typedef.name] = _djson_typedef(typedef)
        end
    end

    local root_def = SchemaTypeDef(typedefs)
    return _SchemaAst.new(root_def)
end


_SchemaAst.TYPE_INTROSPECTION_QUERY = TYPE_INTROSPECTION_QUERY


return _SchemaAst
