-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local random        = require "kong.openid-connect.random"
local codec         = require "kong.openid-connect.codec"
local hash          = require "kong.openid-connect.hash"
local set           = require "kong.openid-connect.set"


local setmetatable  = setmetatable
local decode_args   = codec.args.decode
local encode_args   = codec.args.encode
local get_post_args = ngx.req.get_post_args
local get_uri_args  = ngx.req.get_uri_args
local get_method    = ngx.req.get_method
local read_body     = ngx.req.read_body
local tonumber      = tonumber
local concat        = table.concat
local base64        = codec.base64
local pairs         = pairs
local lower         = string.lower
local find          = string.find
local type          = type
local byte          = string.byte
local pcall         = pcall
local sub           = string.sub
local var           = ngx.var


local SLASH = byte("/")


local OFFLINE_FIX = {
  ["https://accounts.google.com"] = "offline",
}


local authorization = {}


authorization.__index = authorization


function authorization.new(oic)
  return setmetatable({ oic = oic }, authorization)
end


function authorization:bearer(search)
  local opts = self.oic.options

  search = search or opts.auth_methods or { "header", "body", "query" }

  if type(search) ~= "table" then
    search = { search }
  end

  if set.has("header", search) or set.has("client_secret_basic", search) then
    local act = var.http_authorization
    if act then
      local act_type = lower(sub(act, 1, 6))
      if act_type == "bearer" then
        return sub(act, 8), "header"
      end
    end
  end

  if set.has("body", search) or set.has("client_secret_post", search) then
    if find(var.content_type or "", "application/x-www-form-urlencoded", 1, true) == 1 then
      read_body()
      local args = get_post_args()
      if args then
        local act = args.access_token
        if act then
          return act, "body"
        end
      end
    end
  end

  if set.has("query", search) then
    local args = get_uri_args()
    if args then
      local act = args.access_token
      if act then
        return act, "query"
      end
    end
  end
end


function authorization:basic(search)
  local opts = self.oic.options

  search = search or opts.auth_methods or { "header", "body", "query" }

  if type(search) ~= "table" then
    search = { search }
  end

  if set.has("header", search) or set.has("client_secret_basic", search) then
    local act = var.http_authorization
    if act then
      local act_type = lower(sub(act, 1, 5))
      if act_type == "basic" then
        local encoded = sub(act, 7)
        local decoded, err = base64.decode(encoded)
        if decoded then
          local s = find(decoded, ":", 2, true)
          if s then
            local grant_type
            if find(var.content_type or "", "application/x-www-form-urlencoded", 1, true) == 1 then
              read_body()
              local args = get_post_args()
              if args.grant_type == "password" or args.grant_type == "client_credentials" then
                grant_type = args.grant_type
              end
            end

            if not grant_type then
              local args = get_uri_args()
              if args.grant_type == "password" or args.grant_type == "client_credentials" then
                grant_type = args.grant_type
              end
            end

            local identity = sub(decoded, 1, s - 1)
            local secret   = sub(decoded, s + 1)
            return identity, secret, grant_type, "header"
          end
        else
          local s = find(encoded, ":", 2, true)
          if s then
            local grant_type
            if find(var.content_type or "", "application/x-www-form-urlencoded", 1, true) == 1 then
              read_body()
              local args = get_post_args()
              if args.grant_type == "password" or args.grant_type == "client_credentials" then
                grant_type = args.grant_type
              end
            end

            if not grant_type then
              local args = get_uri_args()
              if args.grant_type == "password" or args.grant_type == "client_credentials" then
                grant_type = args.grant_type
              end
            end

            local identity = sub(encoded, 1, s - 1)
            local secret   = sub(encoded, s + 1)
            return identity, secret, grant_type, "header"
          end
        end
        return nil, err
      end
    end
  end

  if set.has("body", search) or set.has("client_secret_post", search) then
    if find(var.content_type or "", "application/x-www-form-urlencoded", 1, true) == 1 then
      read_body()
      local args = get_post_args()
      if args then
        local username = args.username
        if username then
          local password = args.password
          if password then
            return username, password, "password", "body"
          end
        end
        local client_id = args.client_id
        if client_id then
          local client_secret = args.client_secret
          if client_secret then
            return client_id, client_secret, "client_credentials", "body"
          end
        end
      end
    end
  end

  if set.has("query", search) then
    local args = get_uri_args()
    if args then
      local username = args.username
      if username then
        local password = args.password
        if password then
          return username, password, "password", "query"
        end
      end
      local client_id = args.client_id
      if client_id then
        local client_secret = args.client_secret
        if client_secret then
          return client_id, client_secret, "client_credentials", "query"
        end
      end
    end
  end
end


function authorization:state(options)
  options = options or {}

  local opts = self.oic.options
  local args = options.args or opts.args or {}

  local state = options.state or args.state
  if not state then
    local len = tonumber(options.state_len) or tonumber(opts.state_len) or 18
    local hsh = options.state_alg or opts.state_alg
    local cdc = options.state_cdc or opts.state_cdc or "base64url"

    state = random(len, hsh, cdc)
  end
  return state
end


function authorization:nonce(options)
  options = options or {}

  local opts = self.oic.options
  local args = options.args or opts.args or {}

  local nonce = options.nonce or args.nonce
  if not nonce then
    local len = tonumber(options.nonce_len) or tonumber(opts.nonce_len) or 18
    local hsh = options.nonce_alg or opts.nonce_alg
    local cdc = options.nonce_cdc or opts.nonce_cdc or "base64url"

    nonce = random(len, hsh, cdc)
  end
  return nonce
end


function authorization:request(options)
  options = options or {}

  local opts = self.oic.options
  local conf = self.oic.configuration

  local endpoint = options.authorization_endpoint or
                      opts.authorization_endpoint or
                      conf.authorization_endpoint
  if not endpoint then
    return nil, "authorization endpoint was not specified"
  end

  local args = options.args or opts.args or {}
  if type(args) == "string" then
    args = decode_args(args)
  end
  if type(args) ~= "table" then
    return nil, "invalid arguments"
  end

  local verify_parameters = true
  if options.verify_parameters ~= nil then
    verify_parameters = not not options.verify_parameters
  elseif opts.verify_parameters ~= nil then
    verify_parameters = not not opts.verify_parameters
  end

  local response_type, count = set.new(options.response_type or
                                          args.response_type or
                                          opts.response_type or "code")

  if count > 0 then
    if verify_parameters then
      local response_types = options.response_types_supported or
                                opts.response_types_supported or
                                conf.response_types_supported
      if response_types then
        if not set.has(response_type, response_types) then
          return nil, "unsupported response types requested (" .. concat(response_type, ", ") .. ")"
        end
      end
    end

    args.response_type = concat(response_type, " ")
  end

  local scope
  scope, count = set.new(options.scope or
                            args.scope or
                            opts.scope or "openid")

  if count > 0 then
    local scopes = options.scopes_supported or
                      opts.scopes_supported or
                      conf.scopes_supported

    if scopes then
      if not set.has(scope, scopes) then
        if set.has("offline_access", scope) then
          local issuer = conf.issuer
          if issuer then
            if byte(issuer, -1) == SLASH then
              issuer = sub(issuer, 1, -2)
            end

            local access_type = OFFLINE_FIX[issuer]
            if access_type then
              scope = set.remove("offline_access", scope)
              if not args.access_type then
                args.access_type = access_type
              end
            end
          end

          if verify_parameters then
            if not set.has(scope, scopes) then
              return nil, "unsupported scopes requested (" .. concat(scope, ", ") .. ")"
            end
          end
        else
          if verify_parameters then
            return nil, "unsupported scopes requested (" .. concat(scope, ", ") .. ")"
          end
        end
      end
    end

    scope = concat(scope, " ")
    if scope ~= "" then
      args.scope = scope
    end
  end

  local redirect_uri = options.redirect_uri or
                          args.redirect_uri or
                          opts.redirect_uri
  if not redirect_uri then
    local redirect_uris = options.redirect_uris or
                             args.redirect_uris or
                             opts.redirect_uris
    if type(redirect_uris) == "table" and redirect_uris[1] then
      redirect_uri = redirect_uris[1]
    else
      return nil, "redirect uri was not specified"
    end
  end
  args.redirect_uri = redirect_uri

  local client_id = options.client_id or
                       args.client_id or
                       opts.client_id
  if not client_id then
    return nil, "client id was not specified"
  end
  args.client_id = client_id

  local state = self:state(options)
  args.state = state

  local nonce = self:nonce(options)
  args.nonce = nonce

  local code_verifier = options.code_verifier or args.code_verifier

  local pkce = true
  if options.pkce ~= nil then
    pkce = not not options.pkce
  elseif opts.pkce ~= nil then
    pkce = not not opts.pkce
  end

  if pkce then
    local code_challenge_method = options.code_challenge_method or
                                     args.code_challenge_method

    if code_challenge_method then
      if verify_parameters then
        local code_challenge_methods = options.code_challenge_methods_supported or
                                          opts.code_challenge_methods_supported or
                                          conf.code_challenge_methods_supported
        if code_challenge_methods then
          if not set.has(code_challenge_method, code_challenge_methods) then
            return nil, "unsupported code challenge method requested (" .. code_challenge_method .. ")"
          end
        end
      end
    end

    local code_challenge = options.code_challenge or
                              args.code_challenge
    if not code_challenge then
      local code_challenge_methods = options.code_challenge_methods_supported or
                                        opts.code_challenge_methods_supported or
                                        conf.code_challenge_methods_supported
      if code_challenge_methods then
        code_verifier = code_verifier or random(32, nil, "base64url")

        if not code_challenge_method then
          if set.has("S512", code_challenge_methods) then
            code_challenge_method = "S512"

          elseif set.has("S384", code_challenge_methods) then
            code_challenge_method = "S384"

          elseif set.has("S256", code_challenge_methods) then
            code_challenge_method = "S256"

          elseif set.has("plain", code_challenge_methods) then
            code_challenge_method = "plain"

          else
            return nil, "compatible code challenge method was not supported"
          end
        end

        if code_challenge_method == "plain" then
          code_challenge = code_verifier

        else
          code_challenge = codec.base64url.encode(hash[code_challenge_method](code_verifier))
        end
      end
    end

    if code_challenge then
      args.code_challenge        = code_challenge
      args.code_challenge_method = code_challenge_method or "S256"
    end
  end

  local display = options.display or
                     args.display or
                     opts.display
  if display then
    if verify_parameters then
      local displays = options.display_values_supported or
                          opts.display_values_supported or
                          conf.display_values_supported or { "page", "popup", "touch", "wap" }
      if not set.has(display, displays) then
        return nil, "unsupported display value requested (" .. display .. ")"
      end
    end

    args.display = display
  end

  local prompt  = options.prompt or
                     args.prompt or
                     opts.prompt
  if prompt then
    if verify_parameters then
      local prompts = { "none", "login", "consent", "select_account" }
      if not set.has(prompt, prompts) then
        return nil, "unsupported prompt value requested (" .. prompt .. ")"
      end
    end

    args.prompt = prompt
  end

  local acr = options.acr or
                 args.acr or
                 opts.acr
  if acr then
    if verify_parameters then
      local acrs = options.acr_values_supported or
                      opts.acr_values_supported or
                      conf.acr_values_supported
      if acrs then
        if not set.has(acr, acrs) then
          return nil, "unsupported acr value requested (" .. acr .. ")"
        end
      end
    end

    if type(acr) == "table" then
      acr = concat(acr, " ")
    end

    args.acr_values = acr
  end

  local amr = options.amr or
                 args.amr or
                 opts.amr
  if amr then
    if verify_parameters then
      local amrs = options.amr_values_supported or
                      opts.amr_values_supported or
                      conf.amr_values_supported
      if amrs then
        if not set.has(amr, amrs) then
          return nil, "unsupported amr value requested (" .. amr .. ")"
        end
      end
    end

    if type(amr) == "table" then
      amr = concat(amr, " ")
    end

    args.amr_values = amr
  end

  local max_age = options.max_age or
                     args.max_age or
                     opts.max_age
  if max_age then
    max_age = tonumber(max_age)
    if not max_age then
      return nil, "invalid max_age value requested"
    end
    args.max_age = max_age
  end

  local id_token_hint = options.id_token_hint or
                           args.id_token_hint or
                           opts.id_token_hint
  if id_token_hint then
    args.id_token_hint = id_token_hint
  end

  local login_hint = options.login_hint or
                        args.login_hint or
                        opts.login_hint
  if login_hint then
    args.login_hint = login_hint
  end

  local access_type = options.access_type or
                         args.access_type or
                         opts.access_type
  if access_type then
    args.access_type = access_type
  end

  local response_mode = options.response_mode or
                           args.response_mode or
                           opts.response_mode
  if response_mode then
    if verify_parameters then
      local response_modes = options.response_modes_supported or
                                opts.response_modes_supported or
                                conf.response_modes_supported

      if response_modes then
        if not set.has(response_mode, response_modes) then
          return nil, "unsupported response mode requested (" .. response_mode .. ")"
        end
      end
    end

    args.response_mode = response_mode
  end

  local hd = options.hd or args.hd or opts.hd
  if hd then
    args.hd = hd
  end

  local audience = options.audience or
                      args.audience or
                      opts.audience
  if audience then
    if type(audience) == "table" then
      args.audience = concat(audience, " ")

    else
      args.audience = audience
    end
  end

  return {
    url           = endpoint .. "?" .. encode_args(args),
    state         = state,
    nonce         = nonce,
    code_verifier = code_verifier
  }
end


function authorization:verify(options)
  options = options or {}

  local opts = self.oic.options
  local conf = self.oic.configuration
  local args = options.args

  local ok, method = pcall(get_method)
  if ok then
    method = lower(method)

    if method == "post" then
      read_body()
      if not args then
        args = get_post_args()

      else
        local post_args = get_post_args()
        for k, v in pairs(post_args) do
          if not args[k] then
            args[k] = v
          end
        end
      end
    end


    if not args then
      args = get_uri_args()

    else
      local uri_args = get_uri_args()
      for k, v in pairs(uri_args) do
        if not args[k] then
          args[k] = v
        end
      end
    end
  end

  if not args.code then
    return nil, "authorization code not present"
  end

  if not args.state then
    return nil, "authorization state not present"
  end

  if args.state ~= options.state then
    return nil, "invalid authorization state"
  end

  local iss = args.iss
  if iss and conf.issuer and iss ~= conf.issuer then
    return nil, "issuer mismatch"
  end

  if args.client_id then
    local client_id = options.client_id or opts.client_id
    if args.client_id ~= client_id then
      return nil, "client mismatch"
    end
  end

  args.nonce         = options.nonce
  args.code_verifier = options.code_verifier

  if args.id_token or args.access_token then
    options.code = args.code

    local decoded, err = self.oic.token:verify(args, options)
    if not decoded then
      return nil, err
    end

    return args, decoded
  end

  return args
end


return authorization
