-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local padding      = require "kong.openid-connect.padding"
local base16       = require "resty.nettle.base16"
local base64       = require "resty.nettle.base64"
local json         = require "cjson.safe"


local encode_args  = ngx.encode_args
local decode_args  = ngx.decode_args
local escape_uri   = ngx.escape_uri
local unescape_uri = ngx.unescape_uri
local base64_unpad = padding.base64.unpad
local base64_pad   = padding.base64.pad
local find         = string.find
local sub          = string.sub


return {
  json   = {
    encode = json.encode,
    decode = json.decode,
  },
  base16 = {
    encode = base16.encode,
    decode = function(s)
      local err2
      local enc, err = base16.decode(s)
      if not enc then
        enc, err2 = base16.decode("0" .. s)
      end
      if not enc then
        return nil, err or err2
      end
      return enc
    end,
  },
  base64 = {
    encode = function(s)
      local enc, err = base64.encode(s)
      if not enc then
        return nil, err
      end
      enc, err = base64_unpad(enc)
      if not enc then
        return nil, err
      end
      return enc
    end,
    decode = function(s)
      local dec, err = base64_pad(s)
      if not dec then
        return nil, err
      end
      dec, err = base64.decode(dec)
      if not dec then
        return nil, err
      end
      return dec
    end,
  },
  base64url = {
    encode = function(s)
      local enc, err = base64.urlencode(s)
      if not enc then
        return nil, err
      end
      enc, err = base64_unpad(enc)
      if not enc then
        return nil, err
      end
      return enc
    end,
    decode = function(s)
      local dec, err = base64_pad(s)
      if not dec then
        return nil, err
      end
      dec, err = base64.urldecode(dec)
      if not dec then
        return nil, err
      end
      return dec
    end,
  },
  args = {
    encode = encode_args,
    decode = decode_args,
  },
  uri = {
    encode = escape_uri,
    decode = unescape_uri,
  },
  credentials = {
    encode = function(i, s)
      local enc, err = base64.encode(i .. ":" .. s)
      if not enc then
        return nil, err
      end
      return enc
    end,
    decode = function(v)
      local p = find(v, ":", 2, true)
      if not p then
        return nil, "invalid credentials"
      end

      local err

      local i = sub(v, 1, p - 1)
      local s = sub(v, p + 1)

      if i and s then
        i, err = base64_pad(i)
        if not i then
          return nil, err
        end

        i, err = base64.decode(i)
        if not i then
          return nil, err
        end

        s, err = base64_pad(s)
        if not s then
          return nil, err
        end

        s, err = base64.decode()
        if not s then
          return nil, err
        end
      end

      return i, s
    end,
  },
  none = {
    encode = function(s)
      return s
    end,
    decode = function(s)
      return s
    end,
  }
}
