-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local random    = require "kong.openid-connect.random"
local codec     = require "kong.openid-connect.codec"
local rsa       = require "resty.nettle.rsa"
local ecdsa     = require "resty.nettle.ecdsa"
local eddsa     = require "resty.nettle.eddsa"


local ES_CURVES = {
  ES256 = "P-256",
  ES384 = "P-384",
  ES512 = "P-521",
}

local ED_CURVES = {
  EdDSA = "Ed25519",
}

local HSKEYSIZES = {
  HS256 = 32,
  HS384 = 48,
  HS512 = 64,
}


local base64url = codec.base64url
local json      = codec.json


local function encode(key, field, ...)
  local value, err = key[field](key, ...)
  if not value then
    return nil, err
  end

  return base64url.encode(value)
end

local function hs_jwk(jwk)
  jwk = jwk or {
    kty = "oct",
    use = "sig",
    alg = "HS256"
  }

  local ks = HSKEYSIZES[jwk.alg] or 32
  local k, err = base64url.encode(random(ks))
  if not k then
    return nil, err
  end

  local kid
  kid, err = base64url.encode(random(32))
  if not kid then
    return nil, err
  end

  jwk.kty = jwk.kty or "oct"
  jwk.alg = jwk.alg or "HS256"
  jwk.use = jwk.use or "sig"

  jwk.kid = kid
  jwk.k = k

  return jwk
end


local function rs_jwk(jwk)
  local keypair, err = rsa.keypair.new(2048)
  if not keypair then
    return nil, err
  end

  local pub = keypair.public

  local e
  e, err = encode(pub, "e")
  if not e then
    return nil, err
  end

  local n
  n, err = encode(pub, "n")
  if not e then
    return nil, err
  end

  local pri = keypair.private

  local d
  d, err = encode(pri, "d")
  if not d then
    return nil, err
  end

  local p
  p, err = encode(pri, "p")
  if not p then
    return nil, err
  end

  local q
  q, err = encode(pri, "q")
  if not q then
    return nil, err
  end

  local dp
  dp, err = encode(pri, "a")
  if not dp then
    return nil, err
  end

  local dq
  dq, err = encode(pri, "b")
  if not dq then
    return nil, err
  end

  local qi
  qi, err = encode(pri, "c")
  if not qi then
    return nil, err
  end

  local kid
  kid, err = base64url.encode(random(32))
  if not kid then
    return nil, err
  end

  jwk = jwk or {
    kty = "RSA",
    use = "sig",
    alg = "RS256",
  }

  jwk.kty = jwk.kty or "RSA"
  jwk.alg = jwk.alg or "RS256"
  jwk.use = jwk.use or "sig"

  jwk.kid = kid
  jwk.e = e
  jwk.n = n
  jwk.d = d
  jwk.p = p
  jwk.q = q
  jwk.dp = dp
  jwk.dq = dq
  jwk.qi = qi

  return jwk
end


local function es_jwk(jwk)
  jwk = jwk or {
    kty = "EC",
    use = "sig",
  }

  jwk.crv = jwk.crv or ES_CURVES[jwk.alg] or "P-256"

  local keypair, err = ecdsa.keypair.new(jwk.crv)
  if not keypair then
    return nil, err
  end

  local x
  x, err = encode(keypair.point, "x")
  if not x then
    return nil, err
  end

  local y
  y, err = encode(keypair.point, "y")
  if not y then
    return nil, err
  end

  local d
  d, err = encode(keypair.scalar, "d")
  if not d then
    return nil, err
  end

  local kid
  kid, err = base64url.encode(random(32))
  if not kid then
    return nil, err
  end

  jwk.kty = jwk.kty or "EC"
  jwk.use = jwk.use or "sig"

  jwk.kid = kid
  jwk.x = x
  jwk.y = y
  jwk.d = d

  return jwk
end


local function ed_jwk(jwk)
  jwk = jwk or {
    kty = "OKP",
    use = "sig",
    crv = "Ed25519",
  }

  jwk.crv = jwk.crv or ED_CURVES[jwk.alg] or "Ed25519"

  local pri, err = random(32)
  if not pri then
    return nil, err
  end

  local d
  d, err = base64url.encode(pri)
  if not d then
    return nil, err
  end

  local pub
  pub, err = eddsa.public_key(pri)
  if not pub then
    return nil, err
  end

  local x
  x, err = base64url.encode(pub)
  if not x then
    return nil, err
  end

  local kid
  kid, err = base64url.encode(random(32))
  if not kid then
    return nil, err
  end

  jwk.kty = jwk.kty or "OKP"
  jwk.use = jwk.use or "sig"

  jwk.kid = kid
  jwk.x = x
  jwk.d = d

  return jwk
end


local jwks = {}


function jwks.new(options)
  options = options or {}

  local hs256, err = hs_jwk({
    alg = "HS256"
  })

  if not hs256 then
    return nil, err
  end

  local hs384
  hs384, err = hs_jwk({
    alg = "HS384"
  })

  if not hs384 then
    return nil, err
  end

  local hs512
  hs512, err = hs_jwk({
    alg = "HS512"
  })

  if not hs512 then
    return nil, err
  end

  local rs256
  rs256, err = rs_jwk({
    alg = "RS256"
  })

  if not rs256 then
    return nil, err
  end

  local rs512
  rs512, err = rs_jwk({
    alg = "RS512"
  })

  if not rs512 then
    return nil, err
  end

  local ps256
  ps256, err = rs_jwk({
    alg = "PS256"
  })

  if not ps256 then
    return nil, err
  end

  local ps384
  ps384, err = rs_jwk({
    alg = "PS384"
  })

  if not ps384 then
    return nil, err
  end

  local ps512
  ps512, err = rs_jwk({
    alg = "PS512"
  })

  if not ps512 then
    return nil, err
  end

  local es256
  es256, err = es_jwk({
    alg = "ES256"
  })

  if not es256 then
    return nil, err
  end

  local es384
  es384, err = es_jwk({
    alg = "ES384"
  })

  if not es384 then
    return nil, err
  end

  local es512
  es512, err = es_jwk({
    alg = "ES512"
  })

  if not es512 then
    return nil, err
  end

  local ed25519
  ed25519, err = ed_jwk({
    alg = "EdDSA"
  })

  if not eddsa then
    return nil, err
  end

  local keys = {
    hs256,
    hs384,
    hs512,
    rs256,
    rs512,
    ps256,
    ps384,
    ps512,
    es256,
    es384,
    es512,
    ed25519,
  }

  if options.unwrap ~= true then
    keys = {
      keys = keys
    }
  end

  if options.json == true then
    keys, err = json.encode(keys)
    if not keys then
      return nil, err
    end
  end

  return keys
end


return jwks
